@@ -29,6 +29,7 @@
 struct fetch_ctx {
 	unsigned int flags;
 	int outdir_fd, errors;
+	time_t built_after;
 	struct apk_database *db;
 	size_t done, total;
 	struct apk_dependency_array *world;
@@ -69,6 +70,7 @@ static int cup(void)
 }
 
 #define FETCH_OPTIONS(OPT) \
+	OPT(OPT_FETCH_built_after,	APK_OPT_ARG "built-after") \
 	OPT(OPT_FETCH_link,		APK_OPT_SH("l") "link") \
 	OPT(OPT_FETCH_recursive,	APK_OPT_SH("R") "recursive") \
 	OPT(OPT_FETCH_output,		APK_OPT_ARG APK_OPT_SH("o") "output") \
@@ -79,11 +81,30 @@ static int cup(void)
 
 APK_OPT_APPLET(option_desc, FETCH_OPTIONS);
 
+static time_t parse_time(const char *timestr)
+{
+	struct tm tm;
+	char *p;
+	time_t t;
+
+	p = strptime(optarg, "%Y-%m-%d %H:%M:%S", &tm);
+	if (p && *p == 0) return mktime(&tm);
+
+	t = strtoul(optarg, &p, 10);
+	if (p && *p == 0) return t;
+
+	return 0;
+}
+
 static int option_parse_applet(void *ctx, struct apk_db_options *dbopts, int opt, const char *optarg)
 {
 	struct fetch_ctx *fctx = (struct fetch_ctx *) ctx;
 
 	switch (opt) {
+	case OPT_FETCH_built_after:
+		fctx->built_after = parse_time(optarg);
+		if (!fctx->built_after) return -EINVAL;
+		break;
 	case OPT_FETCH_simulate:
 		apk_flags |= APK_SIMULATE;
 		break;
@@ -123,11 +144,9 @@ static void progress_cb(void *pctx, size_t bytes_done)
 	apk_print_progress(ctx->done + bytes_done, ctx->total);
 }
 
-static int fetch_package(apk_hash_item item, void *pctx)
+static int fetch_package(struct apk_database *db, const char *match, struct apk_package *pkg, void *pctx)
 {
-	struct fetch_ctx *ctx = (struct fetch_ctx *) pctx;
-	struct apk_database *db = ctx->db;
-	struct apk_package *pkg = (struct apk_package *) item;
+	struct fetch_ctx *ctx = pctx;
 	struct apk_istream *is;
 	struct apk_repository *repo;
 	struct apk_file_info fi;
@@ -169,10 +188,10 @@ static int fetch_package(apk_hash_item item, void *pctx)
 		fd = STDOUT_FILENO;
 	} else {
 		if ((ctx->flags & FETCH_LINK) && urlfd >= 0) {
-			if (linkat(urlfd, url,
-				   ctx->outdir_fd, filename,
-				   AT_SYMLINK_FOLLOW) == 0)
-				return 0;
+			const char *urlfile = apk_url_local_file(url);
+			if (urlfile &&
+			    linkat(urlfd, urlfile, ctx->outdir_fd, filename, AT_SYMLINK_FOLLOW) == 0)
+				goto done;
 		}
 		fd = openat(ctx->outdir_fd, filename,
 			    O_CREAT|O_RDWR|O_TRUNC|O_CLOEXEC, 0644);
@@ -202,13 +221,13 @@ static int fetch_package(apk_hash_item item, void *pctx)
 		if (r >= 0) r = -EIO;
 		goto err;
 	}
-
-	ctx->done += pkg->size;
-	return 0;
+	goto done;
 
 err:
 	apk_error(PKG_VER_FMT ": %s", PKG_VER_PRINTF(pkg), apk_error_str(r));
 	ctx->errors++;
+done:
+	ctx->done += pkg->size;
 	return 0;
 }
 
@@ -216,6 +235,8 @@ static void mark_package(struct fetch_ctx *ctx, struct apk_package *pkg)
 {
 	if (pkg == NULL || pkg->marked)
 		return;
+	if (ctx->built_after && pkg->build_time && ctx->built_after >= pkg->build_time)
+		return;
 	ctx->total += pkg->size;
 	pkg->marked = 1;
 }
@@ -235,7 +256,7 @@ static void mark_dep_flags(struct fetch_ctx *ctx, struct apk_dependency *dep)
 	apk_deps_add(&ctx->world, dep);
 }
 
-static void mark_name_flags(struct apk_database *db, const char *match, struct apk_name *name, void *pctx)
+static int mark_name_flags(struct apk_database *db, const char *match, struct apk_name *name, void *pctx)
 {
 	struct fetch_ctx *ctx = (struct fetch_ctx *) pctx;
 	struct apk_dependency dep = (struct apk_dependency) {
@@ -244,13 +265,15 @@ static void mark_name_flags(struct apk_database *db, const char *match, struct a
 		.result_mask = APK_DEPMASK_ANY,
 	};
 
-	if (!IS_ERR_OR_NULL(name)) {
-		name->auto_select_virtual = 1;
-		apk_deps_add(&ctx->world, &dep);
-	} else {
+	if (!name) {
 		ctx->errors++;
 		mark_error(ctx, match, name);
+		return 0;
 	}
+
+	name->auto_select_virtual = 1;
+	apk_deps_add(&ctx->world, &dep);
+	return 0;
 }
 
 static void mark_names_recursive(struct apk_database *db, struct apk_string_array *args, void *pctx)
@@ -271,7 +294,7 @@ static void mark_names_recursive(struct apk_database *db, struct apk_string_arra
 	apk_change_array_free(&changeset.changes);
 }
 
-static void mark_name(struct apk_database *db, const char *match, struct apk_name *name, void *ctx)
+static int mark_name(struct apk_database *db, const char *match, struct apk_name *name, void *ctx)
 {
 	struct apk_package *pkg = NULL;
 	struct apk_provider *p;
@@ -284,10 +307,11 @@ static void mark_name(struct apk_database *db, const char *match, struct apk_nam
 
 	if (!pkg) goto err;
 	mark_package(ctx, pkg);
-	return;
+	return 0;
 
 err:
 	mark_error(ctx, match, name);
+	return 0;
 }
 
 static int purge_package(void *pctx, int dirfd, const char *filename)
@@ -346,15 +370,17 @@ static int fetch_main(void *pctx, struct apk_database *db, struct apk_string_arr
 		apk_dependency_array_init(&ctx->world);
 		foreach_array_item(dep, db->world)
 			mark_dep_flags(ctx, dep);
-		apk_name_foreach_matching(db, args, apk_foreach_genid(), mark_name_flags, ctx);
+		if (args->num)
+			apk_db_foreach_matching_name(db, args, mark_name_flags, ctx);
 		if (ctx->errors == 0)
 			mark_names_recursive(db, args, ctx);
 		apk_dependency_array_free(&ctx->world);
 	} else {
-		apk_name_foreach_matching(db, args, apk_foreach_genid(), mark_name, ctx);
+		if (args->num)
+			apk_db_foreach_matching_name(db, args, mark_name, ctx);
 	}
 	if (!ctx->errors)
-		apk_hash_foreach(&db->available.packages, fetch_package, ctx);
+		apk_db_foreach_sorted_package(db, NULL, fetch_package, ctx);
 
 	/* Remove packages not matching download spec from the output directory */
 	if (!ctx->errors && (apk_flags & APK_PURGE) &&
@@ -368,7 +394,7 @@ static struct apk_applet apk_fetch = {
 	.name = "fetch",
 	.open_flags = APK_OPENF_READ | APK_OPENF_NO_STATE,
 	.context_size = sizeof(struct fetch_ctx),
-	.optgroups = { &optgroup_global, &optgroup_applet },
+	.optgroups = { &optgroup_global, &optgroup_source, &optgroup_applet },
 	.main = fetch_main,
 };
 
