@@ -124,20 +124,19 @@ static void progress_cb(void *ctx, size_t installed_bytes)
 			   prog->total.bytes + prog->total.packages);
 }
 
-static int dump_packages(struct apk_changeset *changeset,
+static int dump_packages(struct apk_change_array *changes,
 			 int (*cmp)(struct apk_change *change),
 			 const char *msg)
 {
 	struct apk_change *change;
 	struct apk_name *name;
-	struct apk_indent indent = { .indent = 2 };
+	struct apk_indent indent;
 	int match = 0;
 
-	foreach_array_item(change, changeset->changes) {
-		if (!cmp(change))
-			continue;
-		if (match == 0)
-			printf("%s:\n", msg);
+	apk_print_indented_init(&indent, 0);
+	foreach_array_item(change, changes) {
+		if (!cmp(change)) continue;
+		if (!match) apk_print_indented_group(&indent, 2, "%s:\n", msg);
 		if (change->new_pkg != NULL)
 			name = change->new_pkg->name;
 		else
@@ -146,11 +145,19 @@ static int dump_packages(struct apk_changeset *changeset,
 		apk_print_indented(&indent, APK_BLOB_STR(name->name));
 		match++;
 	}
-	if (match)
-		printf("\n");
+	apk_print_indented_end(&indent);
 	return match;
 }
 
+static int sort_change(const void *a, const void *b)
+{
+	const struct apk_change *ca = a;
+	const struct apk_change *cb = b;
+	const struct apk_name *na = ca->old_pkg ? ca->old_pkg->name : ca->new_pkg->name;
+	const struct apk_name *nb = cb->old_pkg ? cb->old_pkg->name : cb->new_pkg->name;
+	return apk_name_cmp_display(na, nb);
+}
+
 static int cmp_remove(struct apk_change *change)
 {
 	return change->new_pkg == NULL;
@@ -291,30 +298,38 @@ int apk_solver_commit_changeset(struct apk_database *db,
 
 	if ((apk_verbosity > 1 || (apk_flags & APK_INTERACTIVE)) &&
 	    !(apk_flags & APK_SIMULATE)) {
-		r = dump_packages(changeset, cmp_remove,
+		struct apk_change_array *sorted;
+
+		apk_change_array_init(&sorted);
+		apk_change_array_copy(&sorted, changeset->changes);
+		qsort(sorted->item, sorted->num, sizeof(struct apk_change), sort_change);
+
+		r = dump_packages(sorted, cmp_remove,
 				  "The following packages will be REMOVED");
-		r += dump_packages(changeset, cmp_downgrade,
+		r += dump_packages(sorted, cmp_downgrade,
 				   "The following packages will be DOWNGRADED");
 		if (r || (apk_flags & APK_INTERACTIVE) || apk_verbosity > 2) {
-			r += dump_packages(changeset, cmp_new,
+			r += dump_packages(sorted, cmp_new,
 					   "The following NEW packages will be installed");
-			r += dump_packages(changeset, cmp_upgrade,
+			r += dump_packages(sorted, cmp_upgrade,
 					   "The following packages will be upgraded");
-			r += dump_packages(changeset, cmp_reinstall,
+			r += dump_packages(sorted, cmp_reinstall,
 					   "The following packages will be reinstalled");
 			if (download_size) {
 				size_unit = apk_get_human_size(download_size, &humanized);
-				printf("Need to download %lld %s of packages.\n",
+				apk_message("Need to download %lld %s of packages.",
 					(long long)humanized, size_unit);
 			}
 			size_unit = apk_get_human_size(llabs(size_diff), &humanized);
-			printf("After this operation, %lld %s of %s.\n",
+			apk_message("After this operation, %lld %s of %s.",
 				(long long)humanized,
 				size_unit,
 				(size_diff < 0) ?
 				"disk space will be freed" :
 				"additional disk space will be used");
 		}
+		apk_change_array_free(&sorted);
+
 		if (r > 0 && (apk_flags & APK_INTERACTIVE)) {
 			printf("Do you want to continue [Y/n]? ");
 			fflush(stdout);
@@ -403,22 +418,15 @@ struct print_state {
 static void label_start(struct print_state *ps, const char *text)
 {
 	if (ps->label) {
-		printf("  %s:\n", ps->label);
+		apk_print_indented_line(&ps->i, "  %s:\n", ps->label);
 		ps->label = NULL;
-		ps->i.x = ps->i.indent = 0;
 		ps->num_labels++;
 	}
-	if (ps->i.x == 0) {
-		ps->i.x = printf("    %s", text);
-		ps->i.indent = ps->i.x + 1;
-	}
+	if (!ps->i.x) apk_print_indented_group(&ps->i, 0, "    %s", text);
 }
 static void label_end(struct print_state *ps)
 {
-	if (ps->i.x != 0) {
-		printf("\n");
-		ps->i.x = ps->i.indent = 0;
-	}
+	apk_print_indented_end(&ps->i);
 }
 
 static void print_pinning_errors(struct print_state *ps, struct apk_package *pkg, unsigned int tag)
@@ -746,6 +754,7 @@ void apk_solver_print_errors(struct apk_database *db,
 		.db = db,
 		.world = world,
 	};
+	apk_print_indented_init(&ps.i, 1);
 	analyze_deps(&ps, world);
 	foreach_array_item(change, changeset->changes) {
 		struct apk_package *pkg = change->new_pkg;
@@ -754,8 +763,8 @@ void apk_solver_print_errors(struct apk_database *db,
 		analyze_deps(&ps, pkg->depends);
 	}
 
-	if (ps.num_labels == 0)
-		printf("  Huh? Error reporter did not find the broken constraints.\n");
+	if (!ps.num_labels)
+		apk_print_indented_line(&ps.i, "Huh? Error reporter did not find the broken constraints.\n");
 }
 
 int apk_solver_commit(struct apk_database *db,
