@@ -221,10 +221,26 @@ struct apk_name *apk_db_get_name(struct apk_database *db, apk_blob_t name)
 	apk_name_array_init(&pn->rdepends);
 	apk_name_array_init(&pn->rinstall_if);
 	apk_hash_insert_hashed(&db->available.names, pn, hash);
+	db->sorted_names = 0;
 
 	return pn;
 }
 
+static int cmp_provider(const void *a, const void *b)
+{
+	const struct apk_provider *pa = a, *pb = b;
+	return apk_pkg_cmp_display(pa->pkg, pb->pkg);
+}
+
+struct apk_provider_array *apk_name_sorted_providers(struct apk_name *name)
+{
+	if (!name->providers_sorted) {
+		qsort(name->providers->item,  name->providers->num, sizeof name->providers->item[0], cmp_provider);
+		name->providers_sorted = 0;
+	}
+	return name->providers;
+}
+
 static struct apk_db_acl *apk_db_acl_atomize(struct apk_database *db, mode_t mode, uid_t uid, gid_t gid, const struct apk_checksum *xattr_csum)
 {
 	struct apk_db_acl acl = { .mode = mode & 07777, .uid = uid, .gid = gid };
@@ -325,7 +341,7 @@ struct apk_db_dir *apk_db_dir_get(struct apk_database *db, apk_blob_t name)
 	} else if (apk_blob_rsplit(name, '/', &bparent, NULL)) {
 		dir->parent = apk_db_dir_get(db, bparent);
 		dir->protect_mode = dir->parent->protect_mode;
-		dir->has_protected_children = (dir->protect_mode != APK_PROTECT_NONE);
+		dir->has_protected_children = !apk_protect_mode_none(dir->protect_mode);
 		ppaths = dir->parent->protected_paths;
 	} else {
 		dir->parent = apk_db_dir_get(db, APK_BLOB_NULL);
@@ -356,7 +372,7 @@ struct apk_db_dir *apk_db_dir_get(struct apk_database *db, apk_blob_t name)
 
 			dir->protect_mode = ppath->protect_mode;
 		}
-		dir->has_protected_children |= (ppath->protect_mode != APK_PROTECT_NONE);
+		dir->has_protected_children |= !apk_protect_mode_none(ppath->protect_mode);
 	}
 
 	return dir;
@@ -926,7 +942,8 @@ static void apk_blob_push_db_acl(apk_blob_t *b, char field, struct apk_db_acl *a
 static int apk_db_write_fdb(struct apk_database *db, struct apk_ostream *os)
 {
 	struct apk_installed_package *ipkg;
-	struct apk_package *pkg;
+	struct apk_package *pkg, **ppkg;
+	struct apk_package_array *pkgs;
 	struct apk_db_dir_instance *diri;
 	struct apk_db_file *file;
 	struct hlist_node *c1, *c2;
@@ -934,9 +951,11 @@ static int apk_db_write_fdb(struct apk_database *db, struct apk_ostream *os)
 	apk_blob_t bbuf = APK_BLOB_BUF(buf);
 	int r;
 
-	list_for_each_entry(ipkg, &db->installed.packages, installed_pkgs_list) {
-		pkg = ipkg->pkg;
-		r = apk_pkg_write_index_entry(pkg, os);
+	pkgs = apk_db_sorted_installed_packages(db);
+	foreach_array_item(ppkg, pkgs) {
+		pkg = *ppkg;
+		ipkg = pkg->ipkg;
+		r = apk_pkg_write_index_header(pkg, os);
 		if (r < 0)
 			return r;
 
@@ -1161,9 +1180,9 @@ static int apk_db_read_state(struct apk_database *db, int flags)
 	 * 4. scripts db
 	 */
 	if (!(flags & APK_OPENF_NO_WORLD)) {
-		blob = world = apk_blob_from_file(db->root_fd, apk_world_file);
-		if (APK_BLOB_IS_NULL(blob)) return -ENOENT;
-		blob = apk_blob_trim(blob);
+		r = apk_blob_from_file(db->root_fd, apk_world_file, &world);
+		if (r) return r;
+		blob = apk_blob_trim(world);
 		apk_blob_pull_deps(&blob, db, &db->world);
 		free(world.ptr);
 	}
@@ -1184,38 +1203,11 @@ static int apk_db_read_state(struct apk_database *db, int flags)
 	return 0;
 }
 
-struct index_write_ctx {
-	struct apk_ostream *os;
-	int count;
-	int force;
-};
-
-static int write_index_entry(apk_hash_item item, void *ctx)
-{
-	struct index_write_ctx *iwctx = (struct index_write_ctx *) ctx;
-	struct apk_package *pkg = (struct apk_package *) item;
-	int r;
-
-	if (!iwctx->force && pkg->filename == NULL)
-		return 0;
-
-	r = apk_pkg_write_index_entry(pkg, iwctx->os);
-	if (r < 0)
-		return r;
-
-	if (apk_ostream_write(iwctx->os, "\n", 1) != 1)
-		return -EIO;
-
-	iwctx->count++;
-	return 0;
-}
-
 static int apk_db_index_write_nr_cache(struct apk_database *db)
 {
-	struct index_write_ctx ctx = { NULL, 0, TRUE };
-	struct apk_installed_package *ipkg;
+	struct apk_package_array *pkgs;
+	struct apk_package **ppkg;
 	struct apk_ostream *os;
-	int r;
 
 	if (!apk_db_cache_active(db)) return 0;
 
@@ -1227,30 +1219,15 @@ static int apk_db_index_write_nr_cache(struct apk_database *db)
 				 0644);
 	if (IS_ERR_OR_NULL(os)) return PTR_ERR(os);
 
-	ctx.os = os;
-	list_for_each_entry(ipkg, &db->installed.packages, installed_pkgs_list) {
-		struct apk_package *pkg = ipkg->pkg;
+	pkgs = apk_db_sorted_installed_packages(db);
+	foreach_array_item(ppkg, pkgs) {
+		struct apk_package *pkg = *ppkg;
 		if ((pkg->repos == BIT(APK_REPOSITORY_CACHED) ||
 		     (pkg->repos == 0 && !pkg->installed_size))) {
-			r = write_index_entry(pkg, &ctx);
-			if (r != 0) return r;
+			if (apk_pkg_write_index_entry(pkg, os) < 0) break;
 		}
 	}
-	r = apk_ostream_close(os);
-	if (r < 0) return r;
-	return ctx.count;
-}
-
-int apk_db_index_write(struct apk_database *db, struct apk_ostream *os)
-{
-	struct index_write_ctx ctx = { os, 0, FALSE };
-	int r;
-
-	r = apk_hash_foreach(&db->available.packages, write_index_entry, &ctx);
-	if (r < 0)
-		return r;
-
-	return ctx.count;
+	return apk_ostream_close(os);
 }
 
 static int add_protected_path(void *ctx, apk_blob_t blob)
@@ -1266,7 +1243,7 @@ static int add_protected_path(void *ctx, apk_blob_t blob)
 	case '#':
 		return 0;
 	case '-':
-		protect_mode = APK_PROTECT_NONE;
+		protect_mode = APK_PROTECT_IGNORE;
 		break;
 	case '+':
 		protect_mode = APK_PROTECT_CHANGED;
@@ -1315,8 +1292,7 @@ static int add_protected_paths_from_file(void *ctx, int dirfd, const char *file)
 	if (!file_ends_with_dot_list(file))
 		return 0;
 
-	blob = apk_blob_from_file(dirfd, file);
-	if (APK_BLOB_IS_NULL(blob))
+	if (apk_blob_from_file(dirfd, file, &blob))
 		return 0;
 
 	apk_blob_for_each_segment(blob, "\n", add_protected_path, db);
@@ -1365,12 +1341,12 @@ static void handle_alarm(int sig)
 static char *find_mountpoint(int atfd, const char *rel_path)
 {
 	struct mntent *me;
-	struct stat64 st;
+	struct stat st;
 	FILE *f;
 	char *ret = NULL;
 	dev_t dev;
 
-	if (fstatat64(atfd, rel_path, &st, 0) != 0)
+	if (fstatat(atfd, rel_path, &st, 0) != 0)
 		return NULL;
 	dev = st.st_dev;
 
@@ -1380,7 +1356,7 @@ static char *find_mountpoint(int atfd, const char *rel_path)
 	while ((me = getmntent(f)) != NULL) {
 		if (strcmp(me->mnt_fsname, "rootfs") == 0)
 			continue;
-		if (fstatat64(atfd, me->mnt_dir, &st, 0) == 0 &&
+		if (fstatat(atfd, me->mnt_dir, &st, 0) == 0 &&
 		    st.st_dev == dev) {
 			ret = strdup(me->mnt_dir);
 			break;
@@ -1391,7 +1367,7 @@ static char *find_mountpoint(int atfd, const char *rel_path)
 	return ret;
 }
 
-static void mark_in_cache(struct apk_database *db, int dirfd, const char *name, struct apk_package *pkg)
+static void mark_in_cache(struct apk_database *db, int static_cache, int dirfd, const char *name, struct apk_package *pkg)
 {
 	if (pkg == NULL)
 		return;
@@ -1410,8 +1386,7 @@ static int add_repos_from_file(void *ctx, int dirfd, const char *file)
 			return 0;
 	}
 
-	blob = apk_blob_from_file(dirfd, file);
-	if (APK_BLOB_IS_NULL(blob)) {
+	if (apk_blob_from_file(dirfd, file, &blob)) {
 		if (dirfd != AT_FDCWD) return 0;
 		apk_error("failed to read repositories: %s", file);
 		apk_message("NOTE: --repositories-file is relative to the startup directory since apk 2.12.0_rc2");
@@ -1515,6 +1490,8 @@ void apk_db_init(struct apk_database *db)
 	list_init(&db->installed.triggers);
 	apk_dependency_array_init(&db->world);
 	apk_protected_path_array_init(&db->protected_paths);
+	apk_name_array_init(&db->available.sorted_names);
+	apk_package_array_init(&db->installed.sorted_packages);
 	db->permanent = 1;
 	db->root_fd = -1;
 }
@@ -1568,8 +1545,7 @@ int apk_db_open(struct apk_database *db, struct apk_db_options *dbopts)
 		write_arch = TRUE;
 	} else {
 		apk_blob_t arch;
-		arch = apk_blob_from_file(db->root_fd, apk_arch_file);
-		if (!APK_BLOB_IS_NULL(arch)) {
+		if (!apk_blob_from_file(db->root_fd, apk_arch_file, &arch)) {
 			db->arch = apk_atomize_dup(&db->atoms, apk_blob_trim(arch));
 			free(arch.ptr);
 		} else {
@@ -1633,11 +1609,15 @@ int apk_db_open(struct apk_database *db, struct apk_db_options *dbopts)
 		}
 	}
 
-	blob = APK_BLOB_STR("+etc\n" "@etc/init.d\n" "!etc/apk\n");
-	apk_blob_for_each_segment(blob, "\n", add_protected_path, db);
+	if (!APK_BLOB_IS_NULL(dbopts->protected_paths)) {
+		apk_blob_for_each_segment(dbopts->protected_paths, "\n", add_protected_path, db);
+	} else {
+		blob = APK_BLOB_STR("+etc\n" "@etc/init.d\n" "!etc/apk\n");
+		apk_blob_for_each_segment(blob, "\n", add_protected_path, db);
 
-	apk_dir_foreach_file(openat(db->root_fd, "etc/apk/protected_paths.d", O_RDONLY | O_CLOEXEC),
-			     add_protected_paths_from_file, db);
+		apk_dir_foreach_file(openat(db->root_fd, "etc/apk/protected_paths.d", O_RDONLY | O_CLOEXEC),
+				     add_protected_paths_from_file, db);
+	}
 
 	/* figure out where to have the cache */
 	fd = openat(db->root_fd, dbopts->cache_dir, O_RDONLY | O_CLOEXEC);
@@ -1703,10 +1683,12 @@ int apk_db_open(struct apk_database *db, struct apk_db_options *dbopts)
 		}
 	}
 
-	if (!(dbopts->open_flags & APK_OPENF_NO_SYS_REPOS)) {
+	if (!(dbopts->open_flags & APK_OPENF_NO_CMDLINE_REPOS)) {
 		list_for_each_entry(repo, &dbopts->repository_list, list)
 			apk_db_add_repository(db, APK_BLOB_STR(repo->url));
+	}
 
+	if (!(dbopts->open_flags & APK_OPENF_NO_SYS_REPOS)) {
 		if (dbopts->repositories_file == NULL) {
 			add_repos_from_file(db, db->root_fd, "etc/apk/repositories");
 			apk_dir_foreach_file(openat(db->root_fd, "etc/apk/repositories.d", O_RDONLY | O_CLOEXEC),
@@ -1717,12 +1699,12 @@ int apk_db_open(struct apk_database *db, struct apk_db_options *dbopts)
 
 		if (db->repositories.updated > 0)
 			apk_db_index_write_nr_cache(db);
-
-		apk_hash_foreach(&db->available.names, apk_db_name_rdepends, db);
 	}
 
+	apk_hash_foreach(&db->available.names, apk_db_name_rdepends, db);
+
 	if (apk_db_cache_active(db) && (dbopts->open_flags & (APK_OPENF_NO_REPOS|APK_OPENF_NO_INSTALLED)) == 0)
-		apk_db_cache_foreach_item(db, mark_in_cache);
+		apk_db_cache_foreach_item(db, mark_in_cache, 0);
 
 	db->open_complete = 1;
 
@@ -1836,6 +1818,8 @@ void apk_db_close(struct apk_database *db)
 
 	apk_dependency_array_free(&db->world);
 
+	apk_name_array_free(&db->available.sorted_names);
+	apk_package_array_free(&db->installed.sorted_packages);
 	apk_hash_free(&db->available.packages);
 	apk_hash_free(&db->available.names);
 	apk_hash_free(&db->installed.files);
@@ -2038,6 +2022,7 @@ int apk_db_cache_active(struct apk_database *db)
 struct foreach_cache_item_ctx {
 	struct apk_database *db;
 	apk_cache_item_cb cb;
+	int static_cache;
 };
 
 static int foreach_cache_file(void *pctx, int dirfd, const char *name)
@@ -2067,15 +2052,29 @@ static int foreach_cache_file(void *pctx, int dirfd, const char *name)
 		}
 	}
 no_pkg:
-	ctx->cb(db, dirfd, name, pkg);
+	ctx->cb(db, ctx->static_cache, dirfd, name, pkg);
 
 	return 0;
 }
 
-int apk_db_cache_foreach_item(struct apk_database *db, apk_cache_item_cb cb)
+int apk_db_cache_foreach_item(struct apk_database *db, apk_cache_item_cb cb, int static_cache)
 {
-	struct foreach_cache_item_ctx ctx = { db, cb };
+	struct foreach_cache_item_ctx ctx = { db, cb, static_cache };
 
+	if (static_cache) {
+		struct stat st1, st2;
+		int fd = openat(db->root_fd, apk_static_cache_dir, O_RDONLY | O_CLOEXEC);
+		if (fd < 0) return fd;
+		/* Do not handle static cache as static cache if the explicit
+		 * cache is enabled at the static cache location */
+		if (fstat(fd, &st1) == 0 && fstat(db->cache_fd, &st2) == 0 &&
+		    st1.st_dev == st2.st_dev && st1.st_ino == st2.st_ino) {
+			close(fd);
+			return 0;
+		}
+		return apk_dir_foreach_file(fd, foreach_cache_file, &ctx);
+	}
+	if (db->cache_fd < 0) return db->cache_fd;
 	return apk_dir_foreach_file(dup(db->cache_fd), foreach_cache_file, &ctx);
 }
 
@@ -2191,7 +2190,7 @@ static int load_apkindex(void *sctx, const struct apk_file_info *fi,
 	repo = &ctx->db->repos[ctx->repo];
 
 	if (strcmp(fi->name, "DESCRIPTION") == 0) {
-		repo->description = apk_blob_from_istream(is, fi->size);
+		r = apk_blob_from_istream(is, fi->size, &repo->description);
 	} else if (strcmp(fi->name, "APKINDEX") == 0) {
 		ctx->found = 1;
 		r = apk_db_index_read(ctx->db, is, ctx->repo);
@@ -2235,6 +2234,15 @@ int apk_db_index_read_file(struct apk_database *db, const char *file, int repo)
 	return load_index(db, apk_istream_from_file(AT_FDCWD, file), targz, repo);
 }
 
+int apk_db_repository_check(struct apk_database *db)
+{
+	if (apk_force & APK_FORCE_MISSING_REPOSITORIES) return 0;
+	if (!db->repositories.stale && !db->repositories.unavailable) return 0;
+	apk_error("Not continuing due to stale/unavailable repositories."
+		  "Use --force-missing-repositories to continue.");
+	return -1;
+}
+
 int apk_db_add_repository(apk_database_t _db, apk_blob_t _repository)
 {
 	struct apk_database *db = _db.db;
@@ -2440,6 +2448,19 @@ static int contains_control_character(const char *str)
 	return 0;
 }
 
+static int need_checksum(mode_t mode)
+{
+	switch (mode & S_IFMT) {
+	case S_IFSOCK:
+	case S_IFBLK:
+	case S_IFCHR:
+	case S_IFIFO:
+		return FALSE;
+	default:
+		return TRUE;
+	}
+}
+
 static int apk_db_install_archive_entry(void *_ctx,
 					const struct apk_file_info *ae,
 					struct apk_istream *is)
@@ -2629,7 +2650,8 @@ static int apk_db_install_archive_entry(void *_ctx,
 			else
 				memcpy(&file->csum, &ae->csum, sizeof file->csum);
 			/* only warn once per package */
-			if (file->csum.type == APK_CHECKSUM_NONE && !ctx->missing_checksum) {
+			if (file->csum.type == APK_CHECKSUM_NONE &&
+			    need_checksum(ae->mode) && !ctx->missing_checksum) {
 				apk_warning(PKG_VER_FMT": support for packages without embedded "
 					    "checksums will be dropped in apk-tools 3.",
 					    PKG_VER_PRINTF(pkg));
@@ -2692,7 +2714,7 @@ static void apk_db_purge_pkg(struct apk_database *db,
 			};
 			hash = apk_blob_hash_seed(key.filename, diri->dir->hash);
 			if (!is_installed ||
-			    (diri->dir->protect_mode == APK_PROTECT_NONE) ||
+			    apk_protect_mode_none(diri->dir->protect_mode) ||
 			    (apk_flags & APK_PURGE) ||
 			    (file->csum.type != APK_CHECKSUM_NONE &&
 			     apk_fileinfo_get(db->root_fd, name, APK_FI_NOFOLLOW | file->csum.type, &fi, &db->atoms) == 0 &&
@@ -2756,7 +2778,7 @@ static void apk_db_migrate_files(struct apk_database *db,
 				/* File was from overlay, delete the
 				 * packages version */
 				unlinkat(db->root_fd, tmpname, 0);
-			} else if ((diri->dir->protect_mode != APK_PROTECT_NONE) &&
+			} else if (!apk_protect_mode_none(diri->dir->protect_mode) &&
 				   (r == 0) &&
 				   (ofile == NULL ||
 				    ofile->csum.type == APK_CHECKSUM_NONE ||
@@ -2971,76 +2993,251 @@ ret_r:
 struct match_ctx {
 	struct apk_database *db;
 	struct apk_string_array *filter;
-	unsigned int match;
-	void (*cb)(struct apk_database *db, const char *match, struct apk_name *name, void *ctx);
+	apk_db_foreach_name_cb cb;
 	void *cb_ctx;
 };
 
+static int apk_string_match(const char *str, struct apk_string_array *filter, const char **res)
+{
+	char **pmatch;
+
+	foreach_array_item(pmatch, filter) {
+		if (fnmatch(*pmatch, str, FNM_CASEFOLD) == 0) {
+			*res = *pmatch;
+			return 1;
+		}
+	}
+	return 0;
+}
+
+static int apk_name_match(struct apk_name *name, struct apk_string_array *filter, const char **res)
+{
+	if (!filter) {
+		*res = NULL;
+		return 1;
+	}
+	return apk_string_match(name->name, filter, res);
+}
+
+static int apk_pkg_match(struct apk_package *pkg, struct apk_string_array *filter, const char **res, int provides)
+{
+	struct apk_dependency *d;
+
+	if (apk_name_match(pkg->name, filter, res)) return 1;
+	if (!provides) return 0;
+	foreach_array_item(d, pkg->provides) {
+		if (apk_string_match(d->name->name, filter, res)) return 1;
+	}
+	return 0;
+}
+
 static int match_names(apk_hash_item item, void *pctx)
 {
 	struct match_ctx *ctx = (struct match_ctx *) pctx;
 	struct apk_name *name = (struct apk_name *) item;
-	unsigned int genid = ctx->match & APK_FOREACH_GENID_MASK;
-	char **pmatch;
-
-	if (genid) {
-		if (name->foreach_genid >= genid)
-			return 0;
-		name->foreach_genid = genid;
-	}
-
-	if (ctx->filter->num == 0) {
-		ctx->cb(ctx->db, NULL, name, ctx->cb_ctx);
-		return 0;
-	}
-
-	foreach_array_item(pmatch, ctx->filter) {
-		if (fnmatch(*pmatch, name->name, 0) == 0) {
-			ctx->cb(ctx->db, *pmatch, name, ctx->cb_ctx);
-			if (genid)
-				break;
-		}
-	}
+	const char *match;
 
+	if (apk_name_match(name, ctx->filter, &match))
+		return ctx->cb(ctx->db, match, name, ctx->cb_ctx);
 	return 0;
 }
 
-void apk_name_foreach_matching(struct apk_database *db, struct apk_string_array *filter, unsigned int match,
-			       void (*cb)(struct apk_database *db, const char *match, struct apk_name *name, void *ctx),
-			       void *ctx)
+int apk_db_foreach_matching_name(struct apk_database *db, struct apk_string_array *filter,
+				 apk_db_foreach_name_cb cb, void *ctx)
 {
 	char **pmatch;
-	unsigned int genid = match & APK_FOREACH_GENID_MASK;
 	struct apk_name *name;
 	struct match_ctx mctx = {
 		.db = db,
-		.filter = filter,
-		.match = match,
 		.cb = cb,
 		.cb_ctx = ctx,
 	};
+	int r;
 
-	if (filter == NULL || filter->num == 0) {
-		if (!(match & APK_FOREACH_NULL_MATCHES_ALL))
-			return;
-		apk_string_array_init(&mctx.filter);
-		goto all;
-	}
+	if (!filter || !filter->num) goto all;
+
+	mctx.filter = filter;
 	foreach_array_item(pmatch, filter)
 		if (strchr(*pmatch, '*') != NULL)
 			goto all;
 
 	foreach_array_item(pmatch, filter) {
 		name = (struct apk_name *) apk_hash_get(&db->available.names, APK_BLOB_STR(*pmatch));
-		if (genid && name) {
-			if (name->foreach_genid >= genid)
-				continue;
-			name->foreach_genid = genid;
-		}
-		cb(db, *pmatch, name, ctx);
+		r = cb(db, *pmatch, name, ctx);
+		if (r) return r;
 	}
-	return;
+	return 0;
 
 all:
-	apk_hash_foreach(&db->available.names, match_names, &mctx);
+	return apk_hash_foreach(&db->available.names, match_names, &mctx);
+}
+
+static int cmp_name(const void *a, const void *b)
+{
+	const struct apk_name * const* na = a, * const* nb = b;
+	return apk_name_cmp_display(*na, *nb);
+}
+
+static int cmp_package(const void *a, const void *b)
+{
+	const struct apk_package * const* pa = a, * const* pb = b;
+	return apk_pkg_cmp_display(*pa, *pb);
+}
+
+struct add_name_ctx {
+	struct apk_name_array *a;
+	size_t i;
+};
+
+static int add_name(apk_hash_item item, void *ctx)
+{
+	struct apk_name *name = (struct apk_name *) item;
+	struct add_name_ctx *a = ctx;
+	a->a->item[a->i++] = name;
+	return 0;
+}
+
+static struct apk_name_array *apk_db_sorted_names(struct apk_database *db)
+{
+	if (!db->sorted_names) {
+		apk_name_array_resize(&db->available.sorted_names, db->available.names.num_items);
+
+		struct add_name_ctx ctx = { .a = db->available.sorted_names, .i = 0 };
+		apk_hash_foreach(&db->available.names, add_name, &ctx);
+
+		qsort(db->available.sorted_names->item, db->available.sorted_names->num,
+		      sizeof(db->available.sorted_names->item[0]), cmp_name);
+		db->sorted_names = 1;
+	}
+	return db->available.sorted_names;
+}
+
+struct apk_package_array *apk_db_sorted_installed_packages(struct apk_database *db)
+{
+	struct apk_installed_package *ipkg;
+	int n = 0;
+
+	if (!db->sorted_installed_packages) {
+		db->sorted_installed_packages = 1;
+		apk_package_array_resize(&db->installed.sorted_packages, db->installed.stats.packages);
+		list_for_each_entry(ipkg, &db->installed.packages, installed_pkgs_list)
+			db->installed.sorted_packages->item[n++] = ipkg->pkg;
+		qsort(db->installed.sorted_packages->item, db->installed.sorted_packages->num,
+		      sizeof db->installed.sorted_packages->item[0], cmp_package);
+	}
+	return db->installed.sorted_packages;
+}
+
+int apk_db_foreach_sorted_name(struct apk_database *db, struct apk_string_array *filter,
+			       apk_db_foreach_name_cb cb, void *cb_ctx)
+{
+	int walk_all = 0;
+	char **pmatch;
+	const char *match;
+	struct apk_name *name;
+	struct apk_name *results[128], **res;
+	size_t i, num_res = 0;
+	int r;
+
+	if (filter && filter->num) {
+		foreach_array_item(pmatch, filter) {
+			name = (struct apk_name *) apk_hash_get(&db->available.names, APK_BLOB_STR(*pmatch));
+			if (strchr(*pmatch, '*')) {
+				walk_all = 1;
+				continue;
+			}
+			if (!name) {
+				cb(db, *pmatch, NULL, cb_ctx);
+				continue;
+			}
+			if (walk_all) continue;
+			if (num_res >= ARRAY_SIZE(results)) {
+				walk_all = 1;
+				continue;
+			}
+			results[num_res++] = name;
+		}
+	} else {
+		walk_all = 1;
+		filter = NULL;
+	}
+
+	if (walk_all) {
+		struct apk_name_array *a = apk_db_sorted_names(db);
+		res = a->item;
+		num_res = a->num;
+	} else {
+		qsort(results, num_res, sizeof results[0], cmp_name);
+		res = results;
+	}
+
+	for (i = 0; i < num_res; i++) {
+		name = res[i];
+		if (apk_name_match(name, filter, &match)) {
+			r = cb(db, match, name, cb_ctx);
+			if (r) return r;
+		}
+	}
+	return 0;
+}
+
+int __apk_db_foreach_sorted_package(struct apk_database *db, struct apk_string_array *filter,
+				    apk_db_foreach_package_cb cb, void *cb_ctx, int provides)
+{
+	char **pmatch;
+	const char *match;
+	struct apk_name *name;
+	struct apk_package *results[128];
+	struct apk_provider *p;
+	size_t i, num_res = 0;
+	int r;
+
+	if (!filter || !filter->num) {
+		filter = NULL;
+		goto walk_all;
+	}
+
+	foreach_array_item(pmatch, filter) {
+		name = (struct apk_name *) apk_hash_get(&db->available.names, APK_BLOB_STR(*pmatch));
+		if (strchr(*pmatch, '*')) goto walk_all;
+		if (!name) {
+			cb(db, *pmatch, NULL, cb_ctx);
+			continue;
+		}
+
+		foreach_array_item(p, name->providers) {
+			if (!provides && p->pkg->name != name) continue;
+			if (p->pkg->seen) continue;
+			p->pkg->seen = 1;
+			if (num_res >= ARRAY_SIZE(results)) goto walk_all;
+			results[num_res++] = p->pkg;
+		}
+	}
+	for (i = 0; i < num_res; i++) results[i]->seen = 0;
+
+	qsort(results, num_res, sizeof results[0], cmp_package);
+	for (i = 0; i < num_res; i++) {
+		if (apk_pkg_match(results[i], filter, &match, provides)) {
+			r = cb(db, match, results[i], cb_ctx);
+			if (r) return r;
+		}
+	}
+	return 0;
+
+walk_all:
+	for (i = 0; i < num_res; i++) results[i]->seen = 0;
+
+	struct apk_name_array *a = apk_db_sorted_names(db);
+	for (i = 0; i < a->num; i++) {
+		name = a->item[i];
+		apk_name_sorted_providers(name);
+		foreach_array_item(p, name->providers) {
+			if (p->pkg->name != name) continue;
+			if (apk_pkg_match(p->pkg, filter, &match, provides)) {
+				r = cb(db, match, p->pkg, cb_ctx);
+				if (r) return r;
+			}
+		}
+	}
+	return 0;
 }
