@@ -586,55 +586,62 @@ err:
 	return r;
 }
 
-apk_blob_t apk_blob_from_istream(struct apk_istream *is, size_t size)
+int apk_blob_from_istream(struct apk_istream *is, size_t size, apk_blob_t *b)
 {
 	void *ptr;
 	ssize_t rsize;
 
+	*b = APK_BLOB_NULL;
+
 	ptr = malloc(size);
-	if (ptr == NULL)
-		return APK_BLOB_NULL;
+	if (!ptr) return -ENOMEM;
 
 	rsize = apk_istream_read(is, ptr, size);
 	if (rsize < 0) {
 		free(ptr);
-		return APK_BLOB_NULL;
+		return rsize;
 	}
 	if (rsize != size)
 		ptr = realloc(ptr, rsize);
 
-	return APK_BLOB_PTR_LEN(ptr, rsize);
+	*b = APK_BLOB_PTR_LEN(ptr, rsize);
+	return 0;
 }
 
-apk_blob_t apk_blob_from_file(int atfd, const char *file)
+int apk_blob_from_file(int atfd, const char *file, apk_blob_t *b)
 {
-	int fd;
 	struct stat st;
 	char *buf;
+	ssize_t n;
+	int fd;
 
-	if (atfd_error(atfd)) return APK_BLOB_NULL;
+	*b = APK_BLOB_NULL;
+
+	if (atfd_error(atfd)) return atfd;
 
 	fd = openat(atfd, file, O_RDONLY | O_CLOEXEC);
-	if (fd < 0)
-		return APK_BLOB_NULL;
-
-	if (fstat(fd, &st) < 0)
-		goto err_fd;
+	if (fd < 0) goto err;
+	if (fstat(fd, &st) < 0) goto err_fd;
 
 	buf = malloc(st.st_size);
-	if (buf == NULL)
-		goto err_fd;
+	if (!buf) goto err_fd;
 
-	if (read(fd, buf, st.st_size) != st.st_size)
+	n = read(fd, buf, st.st_size);
+	if (n != st.st_size) {
+		if (n >= 0) errno = EIO;
 		goto err_read;
+	}
 
 	close(fd);
-	return APK_BLOB_PTR_LEN(buf, st.st_size);
+	*b = APK_BLOB_PTR_LEN(buf, st.st_size);
+	return 0;
+
 err_read:
 	free(buf);
 err_fd:
 	close(fd);
-	return APK_BLOB_NULL;
+err:
+	return -errno;
 }
 
 int apk_blob_to_file(int atfd, const char *file, apk_blob_t b, unsigned int flags)
@@ -714,7 +721,7 @@ void apk_fileinfo_hash_xattr(struct apk_file_info *fi)
 int apk_fileinfo_get(int atfd, const char *filename, unsigned int flags,
 		     struct apk_file_info *fi, struct apk_atom_pool *atoms)
 {
-	struct stat64 st;
+	struct stat st;
 	unsigned int checksum = flags & 0xff;
 	unsigned int xattr_checksum = (flags >> 8) & 0xff;
 	int atflags = 0;
@@ -725,7 +732,7 @@ int apk_fileinfo_get(int atfd, const char *filename, unsigned int flags,
 	if (flags & APK_FI_NOFOLLOW)
 		atflags |= AT_SYMLINK_NOFOLLOW;
 
-	if (fstatat64(atfd, filename, &st, atflags) != 0)
+	if (fstatat(atfd, filename, &st, atflags) != 0)
 		return -errno;
 
 	*fi = (struct apk_file_info) {
@@ -737,7 +744,7 @@ int apk_fileinfo_get(int atfd, const char *filename, unsigned int flags,
 		.device = st.st_dev,
 	};
 
-	if (xattr_checksum != APK_CHECKSUM_NONE) {
+	if (xattr_checksum != APK_CHECKSUM_NONE && !S_ISLNK(fi->mode)) {
 		ssize_t len, vlen;
 		int fd, i, r;
 		char val[1024], buf[1024];
@@ -775,12 +782,10 @@ int apk_fileinfo_get(int atfd, const char *filename, unsigned int flags,
 
 	/* Checksum file content */
 	if ((flags & APK_FI_NOFOLLOW) && S_ISLNK(st.st_mode)) {
-		char *target = alloca(st.st_size);
-		if (target == NULL)
-			return -ENOMEM;
+		char target[PATH_MAX];
+		if (st.st_size > sizeof target) return -ENOMEM;
 		if (readlinkat(atfd, filename, target, st.st_size) < 0)
 			return -errno;
-
 		EVP_Digest(target, st.st_size, fi->csum.data, NULL,
 			   apk_checksum_evp(checksum), NULL);
 		fi->csum.type = checksum;
