@@ -82,6 +82,7 @@ struct apk_installed_package *apk_pkg_install(struct apk_database *db,
 
 	/* Overlay override information resides in a nameless package */
 	if (pkg->name != NULL) {
+		db->sorted_installed_packages = 0;
 		db->installed.stats.packages++;
 		db->installed.stats.bytes += pkg->installed_size;
 		list_add_tail(&ipkg->installed_pkgs_list,
@@ -598,7 +599,7 @@ int apk_sign_ctx_process_file(struct apk_sign_ctx *ctx,
 	ctx->signature.pkey = PEM_read_bio_PUBKEY(bio, NULL, NULL, NULL);
 	if (ctx->signature.pkey != NULL) {
 		ctx->md = md;
-		ctx->signature.data = apk_blob_from_istream(is, fi->size);
+		apk_blob_from_istream(is, fi->size, &ctx->signature.data);
 	}
 	BIO_free(bio);
 
@@ -957,44 +958,37 @@ err:
 
 void apk_pkg_free(struct apk_package *pkg)
 {
-	if (pkg == NULL)
-		return;
+	if (pkg == NULL) return;
 
 	apk_pkg_uninstall(NULL, pkg);
 	apk_dependency_array_free(&pkg->depends);
 	apk_dependency_array_free(&pkg->provides);
 	apk_dependency_array_free(&pkg->install_if);
-	if (pkg->url)
-		free(pkg->url);
-	if (pkg->description)
-		free(pkg->description);
-	if (pkg->commit)
-		free(pkg->commit);
+	if (pkg->url) free(pkg->url);
+	if (pkg->description) free(pkg->description);
+	if (pkg->commit) free(pkg->commit);
+	if (pkg->filename) free(pkg->filename);
 	free(pkg);
 }
 
+static int apk_ipkg_assign_script(struct apk_installed_package *ipkg, unsigned int type, apk_blob_t b)
+{
+	if (type >= APK_SCRIPT_MAX) {
+		free(b.ptr);
+		return -1;
+	}
+	if (ipkg->script[type].ptr) free(ipkg->script[type].ptr);
+	ipkg->script[type] = b;
+	return 0;
+ }
+
 int apk_ipkg_add_script(struct apk_installed_package *ipkg,
 			struct apk_istream *is,
 			unsigned int type, unsigned int size)
 {
-	void *ptr;
-	int r;
-
-	if (type >= APK_SCRIPT_MAX)
-		return -1;
-
-	ptr = malloc(size);
-	r = apk_istream_read(is, ptr, size);
-	if (r < 0) {
-		free(ptr);
-		return r;
-	}
-
-	if (ipkg->script[type].ptr)
-		free(ipkg->script[type].ptr);
-	ipkg->script[type].ptr = ptr;
-	ipkg->script[type].len = size;
-	return 0;
+	apk_blob_t b;
+	apk_blob_from_istream(is, size, &b);
+	return apk_ipkg_assign_script(ipkg, type, b);
 }
 
 static inline int make_dirs(int root_fd, const char *dirname, mode_t dirmode, mode_t parentmode)
@@ -1114,8 +1108,7 @@ static int write_depends(struct apk_ostream *os, const char *field,
 	return 0;
 }
 
-int apk_pkg_write_index_entry(struct apk_package *info,
-			      struct apk_ostream *os)
+int apk_pkg_write_index_header(struct apk_package *info, struct apk_ostream *os)
 {
 	char buf[2048];
 	apk_blob_t bbuf = APK_BLOB_BUF(buf);
@@ -1178,7 +1171,14 @@ int apk_pkg_write_index_entry(struct apk_package *info,
 	return 0;
 }
 
-int apk_pkg_version_compare(struct apk_package *a, struct apk_package *b)
+int apk_pkg_write_index_entry(struct apk_package *pkg, struct apk_ostream *os)
+{
+	int r = apk_pkg_write_index_header(pkg, os);
+	if (r < 0) return r;
+	return apk_ostream_write(os, "\n", 1);
+}
+
+int apk_pkg_version_compare(const struct apk_package *a, const struct apk_package *b)
 {
 	if (a->version == b->version)
 		return APK_VERSION_EQUAL;
@@ -1186,6 +1186,20 @@ int apk_pkg_version_compare(struct apk_package *a, struct apk_package *b)
 	return apk_version_compare_blob(*a->version, *b->version);
 }
 
+int apk_pkg_cmp_display(const struct apk_package *a, const struct apk_package *b)
+{
+	if (a->name != b->name)
+		return apk_name_cmp_display(a->name, b->name);
+	switch (apk_pkg_version_compare(a, b)) {
+	case APK_VERSION_LESS:
+		return -1;
+	case APK_VERSION_GREATER:
+		return 1;
+	default:
+		return 0;
+	}
+}
+
 unsigned int apk_foreach_genid(void)
 {
 	static unsigned int foreach_genid;
